import torch
from torch import nn


class LabelSmoothing(nn.Module):
    def __init__(self, smoothing=0.05):
        super(LabelSmoothing, self).__init__()
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing

    def forward(self, x, target):
        if self.training:
            x = x.float()
            target = target.float()
            logprobs = torch.nn.functional.log_softmax(x, dim=-1)

            nll_loss = -logprobs * target
            nll_loss = nll_loss.sum(-1)

            smooth_loss = -logprobs.mean(dim=-1)

            loss = self.confidence * nll_loss + self.smoothing * smooth_loss

            return loss.mean()
        else:
            return torch.nn.functional.cross_entropy(x, target)


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count




from torch.utils.data import DataLoader
from torch.optim import optimizer
from .metrics import AverageMeter, RocAucMeter
from torch import nn
import torch
from .TrainerConfig import TrainerConfig
from tqdm import tqdm


class Trainer:

    def __init__(self, model, trainLoader: DataLoader, validLoader: DataLoader, device: torch.device,
                 optim: optimizer, criterion: nn.Module, config: TrainerConfig = None, meters: dict = None):
        self.model = model
        self.trainLoader = trainLoader
        self.validLoader = validLoader
        self.device = device
        self.config = config
        self.optimizer = optim
        self.criterion = criterion.to(self.device)
        self.meters = meters

    def train_one_epoch(self):
        self.model.train()

        summary_loss = AverageMeter.AverageMeter()
        final_scores = RocAucMeter.RocAucMeter()

        for samples, targets in tqdm(self.trainLoader):
            samples = samples.to(self.device).float()
            targets = targets.to(self.device).float()
            batchSize = samples.shape[0]

            self.optimizer.zero_grad()
            outputs = self.model(samples)
            loss = self.criterion(outputs, targets)
            loss.backward()

            final_scores.update(targets, outputs)
            summary_loss.update(loss.detach().item(), batchSize)

            self.optimizer.step()

        return summary_loss, final_scores

    def validation(self):
        self.model.eval()

        summary_loss = AverageMeter.AverageMeter()
        final_scores = RocAucMeter.RocAucMeter()

        for samples, targets in tqdm(self.validLoader):
            with torch.no_grad():
                samples = samples.to(self.device).float()
                targets = targets.to(self.device).float
                batch_size = samples.shape[0]
                outputs = self.model(samples)
                loss = self.criterion(outputs, targets)

                summary_loss.update(loss.detach().item(), batch_size)
                final_scores.update(targets, outputs)

        return summary_loss, final_scores

    def fit(self, epochs: int = 20):
        if self.config is not None and self.config.n_epochs > 0:
            for ep in range(self.config.n_epochs):
                summary_loss, final_scores = self.train_one_epoch()
                print("loss:", summary_loss, "score:", final_scores)
                summary_loss, final_scores = self.validation()
                print("loss:", summary_loss, "score:", final_scores)


import numpy as np
from torch import nn
from sklearn import metrics


def alaska_weighted_auc(y_true, y_valid):
    """
    https://www.kaggle.com/anokas/weighted-auc-metric-updated
    """
    tpr_thresholds = [0.0, 0.4, 1.0]
    weights = [2, 1]

    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_valid, pos_label=1)

    # size of subsets
    areas = np.array(tpr_thresholds[1:]) - np.array(tpr_thresholds[:-1])

    # The total area is normalized by the sum of weights such that the final weighted AUC is between 0 and 1.
    normalization = np.dot(areas, weights)

    competition_metric = 0
    for idx, weight in enumerate(weights):
        y_min = tpr_thresholds[idx]
        y_max = tpr_thresholds[idx + 1]
        mask = (y_min < tpr) & (tpr < y_max)
        # pdb.set_trace()

        x_padding = np.linspace(fpr[mask][-1], 1, 100)

        x = np.concatenate([fpr[mask], x_padding])
        y = np.concatenate([tpr[mask], [y_max] * len(x_padding)])
        y = y - y_min  # normalize such that curve starts at y=0
        score = metrics.auc(x, y)
        submetric = score * weight
        best_subscore = (y_max - y_min) * weight
        competition_metric += submetric

    return competition_metric / normalization


class RocAucMeter(object):
    def __init__(self):
        self.y_true = np.array([0, 1])
        self.y_pred = np.array([0.5, 0.5])
        self.score = 0

    def reset(self):
        self.y_true = np.array([0, 1])
        self.y_pred = np.array([0.5, 0.5])
        self.score = 0

    def update(self, y_true, y_pred):
        y_true = y_true.cpu().numpy().argmax(axis=1).clip(min=0, max=1).astype(int)
        y_pred = 1 - nn.functional.softmax(y_pred, dim=1).data.cpu().numpy()[:, 0]
        self.y_true = np.hstack((self.y_true, y_true))
        self.y_pred = np.hstack((self.y_pred, y_pred))
        self.score = alaska_weighted_auc(self.y_true, self.y_pred)

    @property
    def avg(self):
        return self.score